#ifndef FIX_H
#define FIX_H

#include <common.h>
#include <cstdint>
#include <cmath>
#include <limits>

template <unsigned int s, typename T=int32_t> class Fix
{   
public:
    Fix() : value(0) {}
    Fix(const Fix &other) : value(other.value) { value = other.value; }
    Fix(const float v) { from_float(v); }
    Fix(const double v) { from_float(v); }
    Fix(const int v) { from_int(v); }
    operator float() { return to_float(); }
    operator int() { return to_int(); }
    operator unsigned int() { return to_uint(); }

    T round() const { T ret = value >> (s-1); return (ret>>1) + (ret&1); }

    Fix<s,T> wholes() const { Fix<s,T> ret; ret.value = value & ~((1<<(s+1))-1); return ret; }

    Fix<s,T> operator -() const { Fix<s,T> ret; ret.value = -value; return ret; }
    Fix<s,T>& operator +() { return *this; }

    Fix<s,T> operator +(const int other) const { Fix<s,T> ret; ret.value = value + (other<<s); return ret; }
    Fix<s,T> operator +(const Fix<s,T>& other) const { Fix<s,T> ret; ret.value = value + other.value; return ret; }
    Fix<s,T> operator -(const int other) const { Fix<s,T> ret; ret.value = value - (other<<s); return ret; }
    Fix<s,T> operator -(const Fix<s,T>& other) const { Fix<s,T> ret; ret.value = value - other.value; return ret; }
    Fix<s,T> operator *(const int other) const { Fix<s,T> ret; ret.value = value * other; return ret; }
    Fix<s,T> operator *(const float other) const { Fix<s,T> ret; ret.value = value * other; return ret; }
    Fix<s,T> operator *(const Fix<s,T>& other) const { Fix<s,T> ret; ret.value = (value * other.value) >> s; return ret; }
    Fix<s,T> operator /(const float other) const { Fix<s,T> ret; ret.value = value / other; return ret; }
    Fix<s,T> operator /(const int other) const { Fix<s,T> ret; ret.value = value / other; return ret; }
    Fix<s,T> operator /(const Fix<s,T>& other) const { Fix<s,T> ret; ret.value = (value << s) / other.value; return ret; }

    Fix<s,T>& operator +=(const Fix<s,T>& other) { value += other.value; return *this; }
    Fix<s,T>& operator -=(const Fix<s,T>& other) { value -= other.value; return *this; }
    Fix<s,T>& operator *=(const float other) { value *= other; return *this; }
    Fix<s,T>& operator *=(const int other) { value *= other; return *this; }
    Fix<s,T>& operator *=(const Fix<s,T>& other) { value = (value * other.value) >> s; return *this; }
    Fix<s,T>& operator /=(const float other) { value /= other; return *this; }
    Fix<s,T>& operator /=(const int other) { value /= other; return *this; }
    Fix<s,T>& operator /=(const Fix<s,T>& other) { value = (value << s) / other.value; return *this; }

    bool operator >(const Fix<s,T>& other) const { return value > other.value; }
    bool operator <(const Fix<s,T>& other) const { return value < other.value; }
    bool operator >=(const int other) const { return value >= other<<s; }
    bool operator >=(const Fix<s,T>& other) const { return value >= other.value; }
    bool operator <=(const Fix<s,T>& other) const { return value <= other.value; }
    bool operator ==(const Fix<s,T>& other) const { return value == other.value; }
    bool operator !=(const Fix<s,T>& other) const { return value != other.value; }

    unsigned int to_uint()
    {
        return value >> s;
    }

    int to_int()
    {
        return value >> s;
    }

    float to_float()
    {
        return static_cast<float>(value) / static_cast<float>(1<<s);
    }

    void from_int(const int v)
    {
        value = v << s;
    }

    void from_float(const float v)
    {
        value = static_cast<int32_t>(v * static_cast<float>(1<<s));
    }

    template <unsigned int t>
    operator Fix<t, T>()
    {
        Fix<t, T> ret;
        if(t > s)
            ret.value = value << (t-s);
        else
            ret.value = value >> (s-t);

        return ret;
    }

    Fix<s,T>& normalise_angle()
    {
        if(value < 0)
            value += 360<<s;
        if(value >= (360<<s))
            value -= 360<<s;

        return *this;
    }

    static Fix<s,T> min_step()
    {
        Fix<s,T> ret;
        ret.value = 1;
        return ret;
    }

    static Fix<s,T> max_value()
    {
        Fix<s,T> ret;
        ret.value = std::numeric_limits<T>::max();
        return ret;
    }

    void print()
    {
        char str[32];

        double n = to_float();
        double PRECISION = 0.00000000000001;

        if (n == 0.0)
            strcpy(str, "0");
        else {
            int digit, m, m1;
            char *c = str;
            int neg = (n < 0);
            if (neg)
                n = -n;
            m = log10(n);
            int useExp = (m >= 14 || (neg && m >= 9) || m <= -9);
            if (neg)
                *(c++) = '-';
            if (useExp) {
                if (m < 0)
                   m -= 1.0;
                n = n / pow(10.0, m);
                m1 = m;
                m = 0;}
            if (m < 1.0)
                m = 0;
            while (n > PRECISION || m >= 0) {
                double weight = pow(10.0, m);
                if (weight > 0 && !__builtin_isinf(weight)) {
                    digit = floor(n / weight);
                    n -= (digit * weight);
                    *(c++) = '0' + digit; }
                if (m == 0 && n > 0)
                    *(c++) = '.';
                m--; }
            if (useExp) {
                int i, j;
                *(c++) = 'e';
                if (m1 > 0) {
                    *(c++) = '+';
                } else {
                    *(c++) = '-';
                    m1 = -m1; }
                m = 0;
                while (m1 > 0) {
                    *(c++) = '0' + m1 % 10;
                    m1 /= 10;
                    m++; }
                c -= m;
                for (i = 0, j = m-1; i<j; i++, j--) {
                    c[i] ^= c[j];
                    c[j] ^= c[i];
                    c[i] ^= c[j];}
                c += m;}
            *(c) = '\0'; }

        fputs(str, stdout);
    }

    const static unsigned int precision = s;

    T value;
};

#endif // FIX_H
