#include <os.h>
#include <cmath>
#include <cstdint>
#include <type_traits>
#include <limits>
#include <utility>

#include "gl.h"
#include "fastmath.h"

#define M(m, y, x) (m.data[y][x])
#define P(m, y, x) (m->data[y][x])

MATRIX *transformation;
static COLOR color;
static GLFix u, v;
static COLOR *screen;
static GLFix *z_buffer;
static TEXTURE *texture;
static unsigned int vertices_count = 0;
static VERTEX vertices[4];
static GLDrawMode draw_mode = GL_TRIANGLES;
static bool recording = false;
static VERTEX* record_buffer = nullptr;
static int record_length = 0;
#ifdef SAFE_MODE
    static int matrix_stack_left = MATRIX_STACK_SIZE;
#endif

void nglInit()
{
    init_fastmath();
    transformation = new MATRIX[MATRIX_STACK_SIZE];

    //C++ <3
    z_buffer = new std::remove_reference<decltype(*z_buffer)>::type[WIDTH*HEIGHT];
    glLoadIdentity();
    color = colorRGB(0, 0, 0); //Black
    u = v = 0;

    texture = nullptr;
    vertices_count = 0;
    draw_mode = GL_TRIANGLES;

    recording = false;
    record_buffer = nullptr;
    record_length = 0;

    #ifdef SAFE_MODE
        matrix_stack_left = MATRIX_STACK_SIZE;
    #endif
}

void nglUninit()
{
    uninit_fastmath();
    delete[] transformation;
    delete[] z_buffer;
}

static void mult_mat_mat(MATRIX *mat1, MATRIX *mat2)
{
    GLFix a00 = P(mat1, 0, 0), a01 = P(mat1, 0, 1), a02 = P(mat1, 0, 2), a03 = P(mat1, 0, 3);
    GLFix a10 = P(mat1, 1, 0), a11 = P(mat1, 1, 1), a12 = P(mat1, 1, 2), a13 = P(mat1, 1, 3);
    GLFix a20 = P(mat1, 2, 0), a21 = P(mat1, 2, 1), a22 = P(mat1, 2, 2), a23 = P(mat1, 2, 3);
    GLFix a30 = P(mat1, 3, 0), a31 = P(mat1, 3, 1), a32 = P(mat1, 3, 2), a33 = P(mat1, 3, 3);

    GLFix b00 = P(mat2, 0, 0), b01 = P(mat2, 0, 1), b02 = P(mat2, 0, 2), b03 = P(mat2, 0, 3);
    GLFix b10 = P(mat2, 1, 0), b11 = P(mat2, 1, 1), b12 = P(mat2, 1, 2), b13 = P(mat2, 1, 3);
    GLFix b20 = P(mat2, 2, 0), b21 = P(mat2, 2, 1), b22 = P(mat2, 2, 2), b23 = P(mat2, 2, 3);
    GLFix b30 = P(mat2, 3, 0), b31 = P(mat2, 3, 1), b32 = P(mat2, 3, 2), b33 = P(mat2, 3, 3);

    P(mat1, 0, 0) = a00*b00 + a01*b10 + a02*b20 + a03*b30;
    P(mat1, 0, 1) = a00*b01 + a01*b11 + a02*b21 + a03*b31;
    P(mat1, 0, 2) = a00*b02 + a01*b12 + a02*b22 + a03*b32;
    P(mat1, 0, 3) = a00*b03 + a01*b13 + a02*b23 + a03*b33;
    P(mat1, 1, 0) = a10*b00 + a11*b10 + a12*b20 + a13*b30;
    P(mat1, 1, 1) = a10*b01 + a11*b11 + a12*b21 + a13*b31;
    P(mat1, 1, 2) = a10*b02 + a11*b12 + a12*b22 + a13*b32;
    P(mat1, 1, 3) = a10*b03 + a11*b13 + a12*b23 + a13*b33;
    P(mat1, 2, 0) = a20*b00 + a21*b10 + a22*b20 + a23*b30;
    P(mat1, 2, 1) = a20*b01 + a21*b11 + a22*b21 + a23*b31;
    P(mat1, 2, 2) = a20*b02 + a21*b12 + a22*b22 + a23*b32;
    P(mat1, 2, 3) = a20*b03 + a21*b13 + a22*b23 + a23*b33;
    P(mat1, 3, 0) = a30*b00 + a31*b10 + a32*b20 + a33*b30;
    P(mat1, 3, 1) = a30*b01 + a31*b11 + a32*b21 + a33*b31;
    P(mat1, 3, 2) = a30*b02 + a31*b12 + a32*b22 + a33*b32;
    P(mat1, 3, 3) = a30*b03 + a31*b13 + a32*b23 + a33*b33;
}

static void mult_mat_vect_res(MATRIX *mat1, VERTEX *vect, VERTEX *res)
{
    GLFix x = vect->x, y = vect->y, z = vect->z;

    res->x = P(mat1, 0, 0)*x + P(mat1, 0, 1)*y + P(mat1, 0, 2)*z + P(mat1, 0, 3);
    res->y = P(mat1, 1, 0)*x + P(mat1, 1, 1)*y + P(mat1, 1, 2)*z + P(mat1, 1, 3);
    res->z = P(mat1, 2, 0)*x + P(mat1, 2, 1)*y + P(mat1, 2, 2)*z + P(mat1, 2, 3);
    res->c = vect->c;

    //Virtual center of screen at (0/0)
    res->x -= WIDTH/2;
    res->y -= HEIGHT/2;

    if(res->z == GLFix(0))
        res->z = 1;

    GLFix div = GLFix(NEAR_PLANE)/res->z;

    res->x *= div;
    res->y *= div;

    res->x += WIDTH/2;
    res->y += HEIGHT/2;

    //Screen has y=0 at the top, OpenGL at the bottom
    res->y = GLFix(HEIGHT-1) - res->y;

    //TODO: Move this somewhere else
    #ifdef TEXTURE_SUPPORT
        res->u = vect->u * texture->width;
        res->v = vect->v * texture->height;

        #ifdef SAFE_MODE
            if(res->u > GLFix(texture->width))
            {
                printf("Warning: Texture coord out of bounds!\n");
                res->u = texture->height;
            }
            else if(res->u < GLFix(0))
            {
                printf("Warning: Texture coord out of bounds!\n");
                res->u = 0;
            }

            if(res->v > GLFix(texture->height))
            {
                printf("Warning: Texture coord out of bounds!\n");
                res->v = texture->height;
            }
            else if(res->v < GLFix(0))
            {
                printf("Warning: Texture coord out of bounds!\n");
                res->v = 0;
            }
        #endif

        //We handle texture coords a bit different than OpenGL
        //TODO: Do this better
        if(res->u == GLFix(texture->width))
            res->u = texture->width - 1;

        if(res->v == GLFix(texture->height))
            res->v = texture->height - 1;
    #endif
}

void nglSetBuffer(COLOR *screenBuf)
{
    screen = screenBuf;
}

//Will work only on CX
void nglDisplay()
{
    uint32_t *real_screen = reinterpret_cast<uint32_t*>(SCREEN_BASE_ADDRESS);
    uint32_t *buffer = reinterpret_cast<uint32_t*>(screen);
    for(int x = 0; x < WIDTH*HEIGHT/2; x++)
        *(real_screen++) = *(buffer++);
}

void nglRotateX(GLFix a)
{
    MATRIX rot;

    GLFix sina = fast_sin(a), cosa = fast_cos(a);

    M(rot, 0, 0) = 1;
    M(rot, 1, 1) = cosa;
    M(rot, 1, 2) = -sina;
    M(rot, 2, 1) = sina;
    M(rot, 2, 2) = cosa;
    M(rot, 3, 3) = 1;

    mult_mat_mat(transformation, &rot);
}

void nglRotateY(GLFix a)
{
    MATRIX rot;

    GLFix sina = fast_sin(a), cosa = fast_cos(a);

    M(rot, 0, 0) = cosa;
    M(rot, 0, 2) = sina;
    M(rot, 1, 1) = 1;
    M(rot, 2, 0) = -sina;
    M(rot, 2, 2) = cosa;
    M(rot, 3, 3) = 1;

    mult_mat_mat(transformation, &rot);
}

void nglRotateZ(GLFix a)
{
    MATRIX rot;

    GLFix sina = fast_sin(a), cosa = fast_cos(a);

    M(rot, 0, 0) = cosa;
    M(rot, 0, 1) = -sina;
    M(rot, 1, 0) = sina;
    M(rot, 1, 1) = cosa;
    M(rot, 2, 2) = 1;
    M(rot, 3, 3) = 1;

    mult_mat_mat(transformation, &rot);
}

inline void pixel_pitch(int pitch, GLFix z, COLOR c)
{
    if(z <= GLFix(CLIP_PLANE) || z_buffer[pitch] <= z)
        return;

    z_buffer[pitch] = z;

    #ifdef DRAW_Z_BUFFER
        (void) c;
        screen[pitch] = colorRGB(z / 4000, z / 4000, z / 4000);
    #else
        screen[pitch] = c;
    #endif
}

inline void pixel(int x, int y, GLFix z, COLOR c)
{
    if(x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT)
        return;

    x += y*WIDTH;

    if(z <= GLFix(CLIP_PLANE) || z_buffer[x] <= z)
        return;

    z_buffer[x] = z;

    #ifdef DRAW_Z_BUFFER
        (void) c;
        screen[x] = colorRGB(z / 4000, z / 4000, z / 4000);
    #else
        screen[x] = c;
    #endif
}

RGB rgbColor(COLOR c)
{
    GLFix r = (c >> 11) & 0b11111;
    GLFix g = (c >> 5) & 0b111111;
    GLFix b = (c >> 0) & 0b11111;

    return {r / GLFix(0b11111), g / GLFix(0b111111), b / GLFix(0b11111)};
}

COLOR colorRGB(GLFix r, GLFix g, GLFix b)
{
    int r1 = (r * GLFix(0b11111)).wholes();
    int g1 = (g * GLFix(0b111111)).wholes();
    int b1 = (b * GLFix(0b11111)).wholes();

    return ((r1 & 0b11111) << 11) | ((g1 & 0b111111) << 5) | (b1 & 0b11111);
}

//Doesn't interpolate colors even if enabled
void nglDrawLine3D(VERTEX *v1, VERTEX *v2)
{
    GLFix dy = (v2->y - v1->y) / (v2->x - v1->x);

    COLOR c = v1->c;

    //Height > width? -> Interpolate X
    if(dy > GLFix(1) || dy < GLFix(-1))
    {
        if(v2->y < v1->y)
            std::swap(v1, v2);

        GLFix dx = (v2->x - v1->x) / (v2->y - v1->y);
        GLFix dz = (v2->z - v1->z) / (v2->y - v1->y);

        VERTEX current(*v1);

        if(v2->y >= HEIGHT)
            v2->y = HEIGHT - 1;

        for(; current.y <= v2->y; current.y += GLFix(1))
        {
            pixel(current.x.round(), current.y.round(), current.z, c);

            current.x += dx;
            current.z += dz;
        }
    }
    else
    {
        if(v2->x < v1->x)
            std::swap(v1, v2);

        GLFix dz = (v2->z - v1->z) / (v2->x - v1->x);

        VERTEX current(*v1);

        if(v2->x >= WIDTH)
            v2->x = WIDTH - 1;

        for(; current.x <= v2->x; current.x += GLFix(1))
        {
            pixel(current.x.round(), current.y.round(), current.z, c);

            current.y += dy;
            current.z += dz;
        }
    }
}

/* Basic scanline algorithm, no perspective correct texture mapping.
   In fact, this function is horribly buggy. */
void nglDrawTriangle(VERTEX *low, VERTEX *middle, VERTEX *high)
{
    //I don't know why I have to do this..
    high->x = high->x.wholes();
    high->y = high->y.wholes();

    middle->x = middle->x.wholes();
    middle->y = middle->y.wholes();

    low->x = low->x.wholes();
    low->y = low->y.wholes();

    if(middle->y > high->y)
        std::swap(middle, high);

    if(low->y > middle->y)
        std::swap(low, middle);

    if(middle->y > high->y)
        std::swap(middle, high);

    high->y += 1;

    GLFix height_upper = high->y - middle->y;

    GLFix dx_upper = (high->x - middle->x) / height_upper;
    GLFix dz_upper = (high->z - middle->z) / height_upper;

    GLFix height_lower = middle->y - low->y + 1;

    GLFix dx_lower = (middle->x - low->x) / height_lower;
    GLFix dz_lower = (middle->z - low->z) / height_lower;

    GLFix height_far = high->y - low->y;
    GLFix dx_far = (high->x - low->x) / height_far;
    GLFix dz_far = (high->z - low->z) / height_far;

    #ifdef TEXTURE_SUPPORT
        GLFix du_upper = (high->u - middle->u) / height_upper;
        GLFix dv_upper = (high->v - middle->v) / height_upper;

        GLFix du_lower = (middle->u - low->u) / height_lower;
        GLFix dv_lower = (middle->v - low->v) / height_lower;

        GLFix du_far = (high->u - low->u) / height_far;
        GLFix dv_far = (high->v - low->v) / height_far;

        GLFix ustart = low->u, uend = low->u;
        GLFix vstart = low->v, vend = low->v;
    #elif defined(INTERPOLATE_COLORS)
        RGB high_rgb = rgbColor(high->c);
        RGB middle_rgb = rgbColor(middle->c);
        RGB low_rgb = rgbColor(low->c);

        GLFix dr_upper = (high_rgb.r - middle_rgb.r) / height_upper;
        GLFix dg_upper = (high_rgb.g - middle_rgb.g) / height_upper;
        GLFix db_upper = (high_rgb.b - middle_rgb.b) / height_upper;

        GLFix dr_lower = (middle_rgb.r - low_rgb.r) / height_lower;
        GLFix dg_lower = (middle_rgb.g - low_rgb.g) / height_lower;
        GLFix db_lower = (middle_rgb.b - low_rgb.b) / height_lower;

        GLFix dr_far = (high_rgb.r - low_rgb.r) / height_far;
        GLFix dg_far = (high_rgb.g - low_rgb.g) / height_far;
        GLFix db_far = (high_rgb.b - low_rgb.b) / height_far;

        GLFix rstart = low_rgb.r, rend = low_rgb.r;
        GLFix gstart = low_rgb.g, gend = low_rgb.g;
        GLFix bstart = low_rgb.b, bend = low_rgb.b;
    #endif

    GLFix y = low->y, xstart = low->x, zstart = low->z, xend = low->x, zend = low->z;

    //Vertical clipping
    if(y < GLFix(0))
    {
        GLFix diff = GLFix(0) -y;
        GLFix diff_lower = diff;
        GLFix diff_upper = 0;
        if(diff_lower > height_lower)
        {
            diff_lower = height_lower;
            diff_upper = diff - diff_lower;
        }

        y = 0;

        xstart += dx_far * diff;
        zstart += dz_far * diff;
        xend += dx_lower * diff_lower;
        zend += dz_lower * diff_lower;
        xend += dx_upper * diff_upper;
        zend += dz_upper * diff_upper;

        #ifdef TEXTURE_SUPPORT
            ustart += du_far * diff;
            vstart += dv_far * diff;
            uend += du_lower * diff_lower;
            vend += dv_lower * diff_lower;
            uend += du_upper * diff_upper;
            vend += dv_upper * diff_upper;
        #elif defined(INTERPOLATE_COLORS)
            rstart += dr_far * diff;
            gstart += dg_far * diff;
            bstart += db_far * diff;
            rend = dr_lower * diff_lower;
            gend = dg_lower * diff_lower;
            bend = db_lower * diff_lower;
            rend = dr_upper * diff_upper;
            gend = dg_upper * diff_upper;
            bend = db_upper * diff_upper;
        #endif
    }

    int pitch = y.to_int() * WIDTH;

    GLFix high_y = high->y;

    if(high_y >= GLFix(HEIGHT))
        high_y = HEIGHT - 1;

    //If xstart will get smaller than xend
    if(dx_lower < dx_far)
        goto otherway;

    for(; y <= high_y; y += 1, pitch += WIDTH)
    {
        GLFix line_width = xend - xstart;
        if(line_width == GLFix(0))
            line_width = 1;

        GLFix dz = (zend - zstart) / line_width;
        GLFix z = zstart;

        #ifdef TEXTURE_SUPPORT
            GLFix du = (uend - ustart) / line_width;
            GLFix dv = (vend - vstart) / line_width;
            GLFix u = ustart, v = vstart;
        #elif defined(INTERPOLATE_COLORS)
            GLFix dr = (rend - rstart) / line_width;
            GLFix dg = (gend - gstart) / line_width;
            GLFix db = (bend - bstart) / line_width;

            GLFix r = rstart, g = gstart, b = bstart;
        #endif

        GLFix x1 = xstart;
        GLFix x2 = xend;

        //Horizontal clipping
        if(x1 < GLFix(0))
        {
            GLFix diff = -x1;
            z += dz * diff;

            #ifdef TEXTURE_SUPPORT
                u += du * diff;
                v += dv * diff;
            #elif defined(INTERPOLATE_COLORS)
                r += dr * diff;
                g += dg * diff;
                b += db * diff;
            #endif

            x1 = 0;
        }

        if(x2 >= GLFix(WIDTH))
            x2 = WIDTH - 1;

        for(GLFix x = x1; x <= x2; x += 1)
        {
            #ifdef TEXTURE_SUPPORT
                pixel_pitch(pitch + x.to_int(), z, texture->bitmap[u.to_int() + v.to_int()*texture->width]);
                u += du;
                v += dv;
            #elif defined(INTERPOLATE_COLORS)
                pixel_pitch(pitch + x.to_int(), z, colorRGB(r, g, b));
                r += dr;
                g += dg;
                b += db;
            #else
                pixel_pitch(pitch + x.to_int(), z, low->c);
            #endif

            z += dz;
        }

        if(y <= middle->y)
        {
            xstart += dx_far;
            zstart += dz_far;
            xend += dx_lower;
            zend += dz_lower;

            #ifdef TEXTURE_SUPPORT
                    ustart += du_far;
                    vstart += dv_far;
                    uend += du_lower;
                    vend += dv_lower;
            #elif defined(INTERPOLATE_COLORS)
                    rstart += dr_far;
                    gstart += dg_far;
                    bstart += db_far;

                    rend += dr_lower;
                    gend += dg_lower;
                    bend += db_lower;
            #endif
        }
        else
        {
            xstart += dx_far;
            zstart += dz_far;
            xend += dx_upper;
            zend += dz_upper;

            #ifdef TEXTURE_SUPPORT
                    ustart += du_far;
                    vstart += dv_far;
                    uend += du_upper;
                    vend += dv_upper;
            #elif defined(INTERPOLATE_COLORS)
                    rstart += dr_far;
                    gstart += dg_far;
                    bstart += db_far;

                    rend += dr_upper;
                    gend += dg_upper;
                    bend += db_upper;
            #endif
        }
    }

    return;

    otherway:
    for(; y <= high_y; y += 1, pitch += WIDTH)
    {
        GLFix line_width = xend - xstart;
        if(line_width == GLFix(0))
            line_width = 1;

        //Here come the differences
        GLFix dz = (zend - zstart) / line_width;
        GLFix z = zend;

        #ifdef TEXTURE_SUPPORT
            GLFix du = (uend - ustart) / line_width;
            GLFix dv = (vend - vstart) / line_width;
            GLFix u = uend, v = vend;
        #elif defined(INTERPOLATE_COLORS)
            GLFix dr = (rend - rstart) / line_width;
            GLFix dg = (gend - gstart) / line_width;
            GLFix db = (bend - bstart) / line_width;

            GLFix r = rend, g = gend, b = bend;
        #endif

        GLFix x1 = xend;
        GLFix x2 = xstart;

        //Horizontal clipping
        if(x1 < GLFix(0))
        {
            GLFix diff = -x1;
            z += dz * diff;

            #ifdef TEXTURE_SUPPORT
                u += du * diff;
                v += dv * diff;
            #elif defined(INTERPOLATE_COLORS)
                r += dr * diff;
                g += dg * diff;
                b += db * diff;
            #endif

            x1 = 0;
        }

        if(x2 >= GLFix(WIDTH))
            x2 = WIDTH - 1;

        for(GLFix x = x1; x <= x2; x += 1)
        {
            #ifdef TEXTURE_SUPPORT
                pixel_pitch(pitch + x.to_int(), z, texture->bitmap[u.to_int() + v.to_int()*texture->width]);
                u += du;
                v += dv;
            #elif defined(INTERPOLATE_COLORS)
                pixel_pitch(pitch + x.to_int(), z, colorRGB(r, g, b));
                r += dr;
                g += dg;
                b += db;
            #else
                pixel_pitch(pitch + x.to_int(), z, low->c);
            #endif

            z += dz;
        }

        if(y <= middle->y)
        {
            xstart += dx_far;
            zstart += dz_far;
            xend += dx_lower;
            zend += dz_lower;

            #ifdef TEXTURE_SUPPORT
                    ustart += du_far;
                    vstart += dv_far;
                    uend += du_lower;
                    vend += dv_lower;
            #elif defined(INTERPOLATE_COLORS)
                    rstart += dr_far;
                    gstart += dg_far;
                    bstart += db_far;

                    rend += dr_lower;
                    gend += dg_lower;
                    bend += db_lower;
            #endif
        }
        else
        {
            xstart += dx_far;
            zstart += dz_far;
            xend += dx_upper;
            zend += dz_upper;

            #ifdef TEXTURE_SUPPORT
                    ustart += du_far;
                    vstart += dv_far;
                    uend += du_upper;
                    vend += dv_upper;
            #elif defined(INTERPOLATE_COLORS)
                    rstart += dr_far;
                    gstart += dg_far;
                    bstart += db_far;

                    rend += dr_upper;
                    gend += dg_upper;
                    bend += db_upper;
            #endif
        }
    }

    return;
}

void nglSetColor(COLOR c)
{
    color = c;
}

void nglStartRecording(VERTEX* buffer)
{
    record_buffer = buffer;

    recording = true;
    record_length = 0;
}

int nglStopRecording()
{
    recording = false;

    return record_length;
}

void nglAddVertices(VERTEX *buffer, int length)
{
    while(length--)
        nglAddVertex(buffer++);
}

void nglAddVertex(VERTEX* vertex)
{
    #ifdef SAFE_MODE
        if(texture == nullptr)
        {
            printf("ngl.lang.NoTextureException: Please, don't make me dereference the nullptr!\n");
            return;
        }
    #endif

    if(recording)
    {
        *record_buffer++ = *vertex;
        record_length++;
        return;
    }

    mult_mat_vect_res(transformation, vertex, &vertices[vertices_count]);

    vertices_count++;

    GLFix x1, y1, x2, y2;

    switch(draw_mode)
    {
    case GL_TRIANGLES:
        if(vertices_count != 3)
            break;

        vertices_count = 0;

        //If behind the camera
        if(vertices[0].z <= GLFix(CLIP_PLANE) && vertices[1].z <= GLFix(CLIP_PLANE) && vertices[2].z <= GLFix(CLIP_PLANE))
            break;

        //Backface culling
        x1 = vertices[1].x - vertices[0].x, x2 = vertices[2].x - vertices[0].x;
        y1 = vertices[1].y - vertices[0].y, y2 = vertices[2].y - vertices[0].y;

        if(y1 * x2 >= y2 * x1)
            break;

#ifdef WIREFRAME_MODE
        nglDrawLine3D(&vertecies[0], &vertecies[1]);
        nglDrawLine3D(&vertecies[0], &vertecies[2]);
        nglDrawLine3D(&vertecies[2], &vertecies[1]);
#else
        nglDrawTriangle(&vertices[0], &vertices[1], &vertices[2]);
#endif

        break;

    case GL_QUADS:
        if(vertices_count != 4)
            break;

        vertices_count = 0;

        //If behind the camera
        if(vertices[0].z <= GLFix(CLIP_PLANE) && vertices[1].z <= GLFix(CLIP_PLANE) && vertices[2].z <= GLFix(CLIP_PLANE) && vertices[3].z <= GLFix(CLIP_PLANE))
            break;

        //Backface culling: The quad must be on a plane!
        x1 = vertices[1].x - vertices[0].x, x2 = vertices[2].x - vertices[0].x;
        y1 = vertices[1].y - vertices[0].y, y2 = vertices[2].y - vertices[0].y;

        if(y1 * x2 >= y2 * x1)
            break;

#ifdef WIREFRAME_MODE
        nglDrawLine3D(&vertecies[0], &vertecies[1]);
        nglDrawLine3D(&vertecies[1], &vertecies[2]);
        nglDrawLine3D(&vertecies[2], &vertecies[3]);
        nglDrawLine3D(&vertecies[3], &vertecies[0]);
#else
        nglDrawTriangle(&vertices[0], &vertices[1], &vertices[2]);
        nglDrawTriangle(&vertices[2], &vertices[3], &vertices[0]);
#endif
        break;

    case GL_QUAD_STRIP:
        if(vertices_count != 4)
            break;

        vertices_count = 2;

        if(vertices[0].z <= GLFix(CLIP_PLANE) && vertices[1].z <= GLFix(CLIP_PLANE) && vertices[2].z <= GLFix(CLIP_PLANE) && vertices[3].z <= GLFix(CLIP_PLANE))
            goto end;

        //Backface culling: Each quad must be on a plane!
        x1 = vertices[1].x - vertices[0].x, x2 = vertices[2].x - vertices[0].x;
        y1 = vertices[1].y - vertices[0].y, y2 = vertices[2].y - vertices[0].y;

        if(y1 * x2 >= y2 * x1)
            goto end;

#ifdef WIREFRAME_MODE
        nglDrawLine3D(&vertecies[0], &vertecies[1]);
        nglDrawLine3D(&vertecies[1], &vertecies[2]);
        nglDrawLine3D(&vertecies[2], &vertecies[3]);
        nglDrawLine3D(&vertecies[3], &vertecies[0]);
#else
        nglDrawTriangle(&vertices[0], &vertices[1], &vertices[2]);
        nglDrawTriangle(&vertices[2], &vertices[3], &vertices[0]);
#endif

        end:
        vertices[0] = vertices[2];
        vertices[1] = vertices[3];
        break;
    }
}

void glBindTexture(TEXTURE *tex)
{
    texture = tex;
}

void glColor3f(GLFix r, GLFix g, GLFix b)
{
    color = colorRGB(r, g, b);
}

void glTexCoord2f(GLFix nu, GLFix nv)
{
    u = nu;
    v = nv;
}

void glVertex3f(GLFix x, GLFix y, GLFix z)
{
    VERTEX ver = {
            .x = x,
            .y = y,
            .z = z,
            .u = u,
            .v = v,
            .c = color
    };

    nglAddVertex(&ver);
}

void glBegin(GLDrawMode mode)
{
    vertices_count = 0;
    draw_mode = mode;
}

//Could be faster than memset and allows us to specify a 32-bit value instead of 8-bit
template <typename T>
void memclear(T value, T *buffer, uint32_t size)
{
    while(size--)
        *buffer++ = value;
}

void glClear(int buffers)
{
    if(buffers & GL_COLOR_BUFFER_BIT)
        memclear(color, screen, WIDTH*HEIGHT);

    //The depth buffer uses unsigned 32-bit ints internally
    if(buffers & GL_DEPTH_BUFFER_BIT)
        memclear(z_buffer->max_value(), z_buffer, WIDTH*HEIGHT);
}

void glLoadIdentity()
{
    *transformation = {}; // Copy empty matrix into transformation
    P(transformation, 0, 0) = P(transformation, 1, 1) = P(transformation, 2, 2) = P(transformation, 3, 3) = 1;
}

void glTranslatef(GLFix x, GLFix y, GLFix z)
{
    MATRIX trans;

    M(trans, 0, 0) = M(trans, 1, 1) = M(trans, 2, 2) = M(trans, 3, 3) = 1;
    M(trans, 0, 3) = x;
    M(trans, 1, 3) = y;
    M(trans, 2, 3) = z;

    mult_mat_mat(transformation, &trans);
}

void glScale3f(GLFix x, GLFix y, GLFix z)
{
    MATRIX scale;

    M(scale, 0, 0) = x;
    M(scale, 1, 1) = y;
    M(scale, 2, 2) = z;
    M(scale, 3, 3) = 1;

    mult_mat_mat(transformation, &scale);
}

void glPopMatrix()
{
    #ifdef SAFE_MODE
        if(matrix_stack_left == MATRIX_STACK_SIZE)
        {
            printf("Error: No matrix left on the stack anymore!\n");
            return;
        }
        matrix_stack_left++;
    #endif

    transformation--;
}

void glPushMatrix()
{
    #ifdef SAFE_MODE
        if(matrix_stack_left == 0)
        {
            printf("Error: Matrix stack limit reached!\n");
            return;
        }
        matrix_stack_left--;
    #endif

    transformation++;
    *transformation = *(transformation - 1);
}
